
"""
A lazy domain evaluates its membership function on the fly in terms of that of
other domains.

The `in(x, domain::LazyDomain)` applies three types of transformations:
1. Point mapping: `y = tointernalpoint(domain, x)`
2. Distribution of `y` over member domains given by `elements(domain)`
3. Combination of the outputs into a single boolean result.

The distribution step is determined by the result of `composition(domain)`,
see `composition`. The combination is performed by `combine`. Mapping between
points of the lazy domain and points of its member domains is described by
`y = tointernalpoint(domain,x)` and `x = toexternalpoint(domain, y)`.
"""
abstract type LazyDomain{T} <: Domain{T} end

elements(d::LazyDomain) = d.domains

"""
Supertype of all compositions of a lazy domain. The composition determines how
the point `x` is distributed to the member domains of a lazy domain.

Three compositions implemented in the package are:
- `NoComposition`: the lazy domain encapsulates a single domain and `x` is passed
	through unaltered
- `Combination`: the lazy domain has several members and `x` is passed to the `in`
	method of all members
- `Product`: the lazy domain has several members and the components of `x` are
	passed to the components of the lazy domain
"""
abstract type Composition end

struct NoComposition <: Composition end
struct Combination <: Composition end
struct Product <: Composition end

tointernalpoint(d::LazyDomain, x) = x
toexternalpoint(d::LazyDomain, y) = y

composition(d::LazyDomain) = NoComposition()

indomain(x, d::LazyDomain) = _indomain(tointernalpoint(d, x), d, composition(d), elements(d))
_indomain(x, d, ::NoComposition, domains) = in(x, domains[1])
_indomain(x, d, ::Combination, domains) = combine(d, map(d->in(x, d), domains))
if VERSION >= v"1.2"
	_indomain(x, d, ::Product, domains) = mapreduce(in, &, x, domains)
else
	_indomain(x, d, ::Product, domains) = reduce(&, map(in, x, domains))
end

approx_indomain(x, d::LazyDomain, tolerance) =
	_approx_indomain(tointernalpoint(d, x), d, tolerance, composition(d), elements(d))

_approx_indomain(x, d, tolerance, ::NoComposition, domains) =
    approx_in(x, domains[1], tolerance)
_approx_indomain(x, d, tolerance, ::Combination, domains) =
    combine(d, map(d -> approx_in(x, d, tolerance), domains))
if VERSION >= v"1.2"
	_approx_indomain(x, d, tolerance, ::Product, domains) =
	    mapreduce((u,v)->approx_in(u, v, tolerance), &, x, domains)
else
	_approx_indomain(x, d, tolerance, ::Product, domains) =
	    reduce(&, map((u,v)->approx_in(u, v, tolerance), x, domains))
end

point_in_domain(d::LazyDomain) =
	toexternalpoint(d, map(point_in_domain, elements(d)))

==(a::D, b::D) where {D<:LazyDomain} = elements(a) == elements(b)


const LazyVectorDomain{T} = LazyDomain{Vector{T}}

function dimension(d::LazyVectorDomain)
	dim = dimension(element(d,1))
	@assert all(isequal(dim), map(dimension, elements(d)))
	dim
end

"""
Combine the outputs of `in` of member domains into a single output of the lazy
domain.
"""
combine


"""
DerivedDomain is an abstract supertype for domains that simply wrap another
domain.
"""
abstract type DerivedDomain{T} <: LazyDomain{T} end

elements(d::DerivedDomain) = (d.domain,)

superdomain(d::DerivedDomain) = d.domain

"""
A `WrappedDomain` is a wrapper around an object that implements the domain
interface, and that is itself a domain.
"""
struct WrappedDomain{T,D} <: DerivedDomain{T}
    domain  ::  D
end

WrappedDomain(domain::Domain{T}) where {T} = WrappedDomain{T}(domain)
WrappedDomain(domain) = WrappedDomain{eltype(domain)}(domain)

WrappedDomain{T}(domain::D) where {T,D} = WrappedDomain{T,D}(domain)

convert(::Type{Domain{T}}, d::WrappedDomain) where {T} = WrappedDomain{T}(d.domain)

# Anything can be converted to a domain by wrapping it. An error will be thrown
# if the object does not support `eltype`.
convert(::Type{Domain}, v::Domain) = v
convert(::Type{Domain}, v) = WrappedDomain(v)



"Example of a domain that wraps another domain and thus obtains its own type."
struct ExampleNamedDomain{T,D} <: DerivedDomain{T}
	domain	::	D
end
ExampleNamedDomain(domain::D) where {T,D<:Domain{T}} = ExampleNamedDomain{T,D}(domain)
